@@ -204,9 +204,7 @@ class Bsm(opt.OptAnalyticABC):
 
         # Exclude optoin price below intrinsic value or above max value (1 for call or k for put)
         # ind_solve can be scalar or array. scalar can be fine in np.abs(p_err[ind_solve])
-        ind_solve = (price_std - p_min > Bsm.IMPVOL_TOL) & (
-                p_max - price_std > Bsm.IMPVOL_TOL
-        )
+        ind_solve = (price_std - p_min > Bsm.IMPVOL_TOL) & (p_max - price_std > Bsm.IMPVOL_TOL)
 
         # initial guess = inflection point in sigma (volga=0)
         _sigma = np.ones_like(ind_solve)*np.sqrt(
@@ -287,21 +285,11 @@ class Bsm(opt.OptAnalyticABC):
             kk = strike/fwd
             lnk = np.log(kk)
             if model.lower() == "norm-approx":
-                return (
-                        self.sigma
-                        *fwd
-                        *np.sqrt(kk)
-                        *(1 + lnk**2/24)
-                        /(1 + self.sigma**2*texp/24)
-                )
+                return self.sigma * fwd * np.sqrt(kk) * (1 + lnk**2/24) / (1 + self.sigma**2*texp/24)
             else:
                 with np.errstate(divide="ignore", invalid="ignore"):
                     term1 = np.where(np.fabs(lnk) > 1e-8, (kk - 1)/lnk, 2/(3 - kk))
-                    term2 = np.where(
-                        np.fabs(lnk) > 1e-8,
-                        (np.log(term1) - lnk/2)/lnk**2,
-                        1/24,
-                    )
+                    term2 = np.where(np.fabs(lnk) > 1e-8, (np.log(term1) - lnk/2)/lnk**2, 1/24)
                 return self.sigma*fwd*term1*(1 - term2*self.sigma**2*texp)
         else:
             raise ValueError(f"Unknown model: {model}")
@@ -418,7 +406,7 @@ class Bsm(opt.OptAnalyticABC):
 
 class BsmDisp(smile.OptSmileABC, Bsm):
     """
-    Displaced Black-Scholes-Merton model for option pricing. Displace price,
+    Displaced Black-Scholes-Merton model for option pricing. Displaced price,
 
         D(F_t) = beta*F_t + (1-beta)*A
 
@@ -461,7 +449,7 @@ class BsmDisp(smile.OptSmileABC, Bsm):
 
     @property
     def sigma(self):
-        return self.sigma_disp*self.beta
+        return self.sigma_disp * self.beta
 
     @sigma.setter
     def sigma(self, sigma):
@@ -475,7 +463,7 @@ class BsmDisp(smile.OptSmileABC, Bsm):
             spot: spot (or forward) price
 
         Returns:
-            Displaces spot
+            Displaced spot
         """
         return self.beta*spot + (1 - self.beta)*self.pivot
 
@@ -488,7 +476,7 @@ class BsmDisp(smile.OptSmileABC, Bsm):
             texp: time to expiry
 
         Returns:
-            Displace strike
+            Displaced strike
         """
         return self.beta*strike + (1 - self.beta)*self.forward(self.pivot, texp)
 
@@ -552,13 +540,7 @@ class BsmDisp(smile.OptSmileABC, Bsm):
             kkd = self.disp_strike(strike, texp)/fwdd
             lnkd = np.log(kkd)
             # self.sigma actually means self.beta * self._sigma
-            vol = (
-                    self.sigma_disp
-                    *fwdd
-                    *np.sqrt(kkd)
-                    *(1 + lnkd**2/24)
-                    /(1 + self.sigma**2*texp/24)
-            )
+            vol = self.sigma_disp * fwdd * np.sqrt(kkd) * (1 + lnkd**2/24) / (1 + self.sigma**2*texp/24)
         elif model.lower() == "bsm-approx":
             fwd = self.forward(spot, texp)
             kk = strike/fwd
@@ -570,12 +552,7 @@ class BsmDisp(smile.OptSmileABC, Bsm):
 
             # self.sigma actually means self.beta * self.sigma_disp
             vol = self.sigma_disp*(fwdd/fwd)*np.sqrt(kkd/kk)
-            vol *= (
-                    (1 + lnkd**2/24)
-                    /(1 + lnk**2/24)
-                    *(1 + vol**2*texp/24)
-                    /(1 + self.sigma**2*texp/24)
-            )
+            vol *= (1 + lnkd**2/24) / (1 + lnk**2/24) * (1 + vol**2*texp/24) / (1 + self.sigma**2*texp/24)
         else:
             vol = super().vol_smile(strike, spot, texp, model=model, cp=cp)
 
